class AngularMomentumNotConservedException(Exception):
    def __init__(self, msg):
        super().__init__(msg)

class BodyCollisionException(Exception):
    def __init__(self, msg):
        super().__init__(msg)

class BodyEscapeException(Exception):
    def __init__(self, msg):
        super().__init__(msg)

class COMNotConservedException(Exception):
    def __init__(self, msg):
        super().__init__(msg)

class EnergyNotConservedException(Exception):
    def __init__(self, msg):
        super().__init__(msg)

class LinearMomentumNotConservedException(Exception):
    def __init__(self, msg):
        super().__init__(msg)

class SmallAdaptiveDeltaException(Exception):
    def __init__(self, msg):
        super().__init__(msg)

class Figure8InitException(Exception):
    def __init__(self, msg):
        super().__init__(msg)

def check_exception(condition, exception, msg = "An exception occurred!"):
    if not condition:
        raise exception(msg)


import os
from multiprocessing import Pool
import tqdm

from nbodysim.three_body import *
from nbodysim.exceptions import *
from nbodysim.integrators.leapfrog_3 import Leapfrog3
from nbodysim.stability_investigator.stability_plotter import StabilityPlotter

class MPStabilityPlotter(StabilityPlotter):
    """
    Inherits from StabilityPlotter. Performs the same task, albeit using multiprocessing to speed up the process.
    """
    def __init__(self, perturb, n_trials, collision_tolerance, escape_tolerance, centre_x = 0, centre_y = 0, steps = 10**4, delta = 10**-2, tolerance = 10**-2, adaptive_constant = 0.1, delta_lim = 10**-5):
        """
        :param perturb: the unit amount by which v_x and v_y can be perturbed.
                        That is,any perturbation to vx or vy is a multiple of perturb
        :param n_trials: the number of times that the initial conditions are perturbed to one side of 0.
                         For example, if n_trials = 5, then we will produce a 11 x 11 matrix.
        :param collision_tolerance: maximum distance between 2 bodies allowed before ending simulation
                                    If None, collisions are not considered.
        :param escape_tolerance: maximum distance away from the centre of mass (COM) allowed before ending simulation
                                 If None, escape_tolerance is automatically calculated
                                 If -1, escape_tolerance is not considered
        :param centre_x: the x coordinate of the centre of the stability plot
        :param centre_y: the y coordinate of the centre of the stability plot
        :param steps: the number of steps to integrate for.
                      Since adaptive time step will be used, this is simply used alongside delta to calculate simulation time
        :param delta: time step to use for the integrator. Smaller time steps lead to more accurate orbits.
                      Since adaptive time step will be used, this is simply used alongside steps to calculate simulation time
        :param tolerance: allowed absolute error for determining conservation of calculated quantities.
        :param adaptive_constant: constant resizing factor for adaptive time step
        :param delta_lim: smallest value allowed for the adaptive time step
        """
        super().__init__(perturb, n_trials, collision_tolerance, escape_tolerance, centre_x, centre_y, steps, delta, tolerance, adaptive_constant, delta_lim)

    def get_stability_cell(self, coords):
        """
        Computes the error score for a single cell in the stability matrix
        :param coords: the row and column of the cell (with coords = (row, col)). Used to calculate the perturbations.
        :return: the error score for the NBody instance in coords of the stability matrix
        """

        # extract row and column, in order to calculate perturbation
        row, col = coords

        # the maximum allowed perturbation
        n = self.perturb * self.n_trials

        # the amount by which y component of velocity is changed
        dvy = n - row * self.perturb

        # the amount by which x component of velocity is changed
        dvx = -n + col * self.perturb

        try:
            # check for potential exceptions (either Figure of 8 or adaptive delta) during initialisation
            nbody = get_figure_8(-0.5 * np.array([-0.93240737, -0.86473146, 0]) + np.array([dvx + self.centre_x, dvy + self.centre_y, 0]), -0.24308753,
                                 collision_tolerance=self.collision_tolerance, escape_tolerance=self.escape_tolerance)

            integrator = Leapfrog3(nbody, steps=self.steps, delta=self.delta, tolerance=self.tolerance, adaptive=True,
                                   adaptive_constant=self.adaptive_constant, store_properties=False, delta_lim=self.delta_lim)

            # integrate, and catch any exception in the process
            try:
                integrator.get_orbits()

                # if integration took 10^5 steps (max allowed) return 1. Otherwise, 0.
                if integrator.full_run:
                    return 1
                else:
                    return 0

            except(SmallAdaptiveDeltaException,
                   COMNotConservedException,
                   LinearMomentumNotConservedException,
                   BodyEscapeException,
                   BodyCollisionException,
                   AngularMomentumNotConservedException,
                   EnergyNotConservedException) as e:

                   # return error score (caused during simulation)
                   return self.get_error_score(e, integrator.times[-1])
        except (Figure8InitException,
                SmallAdaptiveDeltaException) as e:

                # return error score (caused during initialisation)
                return self.exception_dict[e.__class__.__name__]

    def get_stability_matrix(self):
        """
        Perturbs Figure 8 orbit, producing a matrix outlining stability regions.
        Uses multiprocessing to compute the stability_matrix.
        :return: a (2*n_trials + 1 x 2*n_trials + 1), with numbers representing errors dependent on how the simulation ended.
                 Error-to-number conversion depedns on exception_dict
        """

        # list containing the arguments to be used when calculating the stability values
        args = []

        # size of the stability_matrix
        n = 2 * self.n_trials + 1

        # compute the arguments in args
        for i in range(n):
            for j in range(n):
                args.append((i, j))

        # calculate number of CPU cores that can be used
        n_cpus = os.cpu_count()

        # use pooling to multiprocess across a list of different arguments
        # use tqdm in order to get progress in calculations
        with Pool(n_cpus) as pool:
            results = list(tqdm.tqdm(pool.imap(func = self.get_stability_cell, iterable = args), total = n*n))

        # results is a 1D array, so reshape into required matrix
        stability_matrix = np.array(results).reshape((n,n))

        return stability_matrix


import numpy as np
import warnings

from nbodysim.exceptions import *

# Helper file containing useful mathematical functions for calculations

def ten_norm(ten, axis = 1, sqrt = True):
    """
    Calculates the norm of a tensor, as defined by the norm of each vector component
    The norm being the square root of the sum of the squares of each component of the tensor
    :param ten: tensor to calculate norm
    :param axis: the "axis" over which the the norm is calculated. For an n-dimensional tensor, use norm n - 1
    :param sqrt: boolean to indicate whether we return the standard norm (with a square root) or not.
    :return: a lower dimensional (n-1) tensor with the norm calculated
    """

    if sqrt:
        return np.sqrt(np.sum(ten**2, axis = axis))
    else:
        return np.sum(ten**2, axis = axis)

def vec_cross(vec1, vec2):
    """
    Calculates cross product between 2 vectors
    """

    i = vec1[1]*vec2[2] - vec1[2]*vec2[1]
    j = vec1[2] * vec2[0] - vec1[0] * vec2[2]
    k = vec1[0]*vec2[1] - vec1[1]*vec2[0]

    cross = np.array([i,j,k])

    return cross

def mat_cross(mat1, mat2):
    """
    Calculates the cross product between 2 (n x 3) matrices,
    Defined as the pairwise cross product of the row vectors of the matrices
    :return: an (n x 3) matrix, with each row being the pairwise cross product of mat1 and mat2
    """

    n = len(mat1)
    mat_cross = np.zeros(shape = (n, 3))

    for i in range(n):
        mat_cross[i] = vec_cross(mat1[i], mat2[i])

    return mat_cross

def perc_change(initial, values, perc = False, init_val = 1):
    """
    Calculates absolute percentage/decimal change between the initial value and either a single value or a series of values
    :param initial: either a single value, or an array of values
    :param values: the values to be compared against initial to calculate percentage change
    :param perc: if True, the change is given as a percentage; otherwise as a decimal
    :param init_val: initial value when calculating percentages, should initial be 0/array of 0s
    :return: the percentage (as decimal or percentage) change between values and initial, in absolute terms
    """

    # check if initial is an array of values
    if isinstance(initial, np.ndarray) and (initial == 0).all():
        warnings.warn(f"The initial quantity was all 0, so percentage change is meaningless!. Comparing with baseline value: {init_val}",
                          category = RuntimeWarning)
        initial = np.ones_like(initial)*init_val
    elif (initial == 0):
        warnings.warn(f"The initial quantity was all 0, so percentage change is meaningless!. Comparing with baseline value: {init_val}",
                category=RuntimeWarning)
        initial = init_val

    change = np.abs((values - initial) / initial)

    # calculate percentage if required
    if perc:
        return change * 100

    return change

def variable_delta(positions, velocities, adaptive_constant, delta_lim =10 ** -5):
    """
    Calculates variable delta for the system
    :param positions: positions of bodies in the system
    :param velocities: velocities of bodies in the system
    :param adaptive_constant: constant resizing factor for variable delta
    :param delta_lim: smallest value allowed for the variable delta
    :return: the calculated variable delta for the system
    """

    # check: positions and velocities are of the same size
    n = len(positions)

    assert n == len(velocities)

    # if only 1 body, then variable delta is the ratio of position magnitude to velocity magnitude, multiplied by adaptive_constant
    if n == 1:
        delta_x = ten_norm(positions[0], sqrt = True, axis = 0)
        delta_v = ten_norm(velocities[0], sqrt=True, axis=0)
        return adaptive_constant * delta_x / delta_v

    # create an (n x n) array to hold all calculated deltas for the system
    potential_deltas = np.ones(shape = (n, n))*np.inf

    # fill potential deltas with the ratio of position magnitude to velocity magnitude
    for i in range(n):
        for j in range(i):
                delta_x = ten_norm((positions[i] - positions[j]), sqrt = True, axis = 0)
                delta_v = ten_norm((velocities[i] - velocities[j]), sqrt = True, axis = 0)
                # entry (i,j) will be the same as (j,i) as we just consider magnitudes
                potential_deltas[i, j] = np.divide(delta_x, delta_v, where = delta_v != 0)
                potential_deltas[j, i] = potential_deltas[i, j]

    # variable delta will be the smallest delta within potential_deltas, multiplied by adaptive_constant
    variable_delta = adaptive_constant * np.amin(potential_deltas)

    # check; if delta is provided, ensure that the calculated variable delta does not become smaller than the minimum allowed
    if delta_lim is not None:
        check_exception(variable_delta > delta_lim, SmallAdaptiveDeltaException, f"Adaptive delta was made too small ({variable_delta}) - orbit unfeasible")

    return variable_delta

def get_relative_normalised_positions(positions):
    # get the relative positions of the bodies
    R_1 = positions[0,:,:] - positions[1,:,:]
    R_2 = positions[1,:,:] - positions[2,:,:]
    R_3 = positions[0,:,:] - positions[2,:,:]

    # get scaling factor
    norm_R_1 = ten_norm(R_1, axis = 1, sqrt = True)
    norm_R_2 = ten_norm(R_2, axis = 1, sqrt = True)
    norm_R_3 = ten_norm(R_3, axis = 1, sqrt = True)
    N = norm_R_1 + norm_R_2 + norm_R_3

    # get the new coordinates
    X_1 = norm_R_1 / N
    X_2 = norm_R_2 / N

    return X_1, X_2

def pair_encoder(x, y):
    return 0.5*(x + y)*(x + y + 1) + y




import numpy as np

from nbodysim import nmath as nm
from nbodysim.nbody import NBody
from nbodysim.exceptions import Figure8InitException, check_exception

# ------------------------------ EULER ------------------------------

def get_euler_nbod(r_1: np.array, masses, G = 1):
    """
    Produces an NBody instance with velocities satisfying the Euler 3-Body configuration
    :param r_1: the position of one of the non-central bodies in the configuration
    :param masses: the masses for each of the 3 bodies
    :param G: constant of gravitation
    :return: an NBody instance with velocities satisfying the Euler 3-Body configuration
    """

    # check: position is a vector in R^3 with the y component being non-zero
    assert r_1[1] != 0 and r_1.shape == (3,)
    assert r_1[2] == 0

    # check: side masses are the same
    assert masses[1] == masses[2]

    # calculate the velocity magnitude from the positions
    velocity_mag = np.sqrt((4*G*masses[0] + G*masses[2]) / (4 * nm.ten_norm(r_1, axis = 0, sqrt = True)))
    print(velocity_mag)

    # initialise positions, with a body at the origin, and the other 2 bodies equally distanced and collinear
    init_positions = np.array([np.zeros(shape=(3,)), r_1, -r_1])

    # create the velocity array, by making the velocity vector perpendicular to the position
    init_velocity = velocity_mag * np.array([r_1[1], -r_1[0], 0]) / nm.ten_norm(r_1, axis=0, sqrt = True)

    # initialise velocities for simulation
    init_velocities = np.array([[0,0,0], init_velocity, -init_velocity])

    return NBody(init_positions, init_velocities, masses, escape_tolerance=-1)

# ------------------------------ LAGRANGE ------------------------------

def rot_mat(theta):
    """
    A rotation matrix, by angle theta (rad)
    """
    return np.array([[np.cos(theta), np.sin(theta), 0],[-np.sin(theta), np.cos(theta), 0], [0, 0, 1]])

def get_lagrange_nbod(r_1: np.array, mass, G = 1):
    """
    Produces an NBody instance with velocities satisfying the Lagrange 3-Body configuration
    :param r_1: the position of any of the 3 bodies
    :param masses: the mass of each of the 3 bodies for each of the 3 bodies
    :param G: constant of gravitation
    :return: an NBody instance with velocities satisfying the Euler 3-Body configuration
    """

    assert r_1[2] == 0

    # initialise rotation matrix, and calculate position vector for the other 2 bodies
    R = rot_mat(2*np.pi/3)
    r_2 = R @ r_1
    r_3 = R @ r_2

    # calculate the gravitational force exerted on the body
    gravitational_force = G * ((mass/nm.ten_norm(r_2 - r_1, axis = 0, sqrt = False)**(1.5))*(r_2 - r_1) +
                               (mass/nm.ten_norm(r_3 - r_1, axis = 0, sqrt = False)**(1.5))*(r_3 - r_1))

    # pseudo-magnitude of centripetal force. Used to obtain the ratio (v/r)^2
    centripetal_mag = -1 * np.divide(gravitational_force, r_1, out=np.zeros_like(gravitational_force), where = r_1 != 0)

    # calculate the magnitude of the velocity vector
    velocity_mag = np.sqrt(centripetal_mag[np.nonzero(centripetal_mag)[0][0]] * nm.ten_norm(r_1, axis=0, sqrt=False))

    # calculate velocity, as a vector perpendicular to r_1
    v_1 = velocity_mag * np.array([r_1[1], -r_1[0], 0])/nm.ten_norm(r_1, axis = 0, sqrt=True)

    # initialise NBody
    init_positions = np.array([r_1, r_2, r_3])
    init_velocities = np.array([v_1, R @ v_1, R @ (R @ v_1)])

    return NBody(init_positions, init_velocities, mass*np.ones(shape = (3,)), escape_tolerance=-1)

# ------------------------------ FIGURE 8 ------------------------------

# initial conditions taken from http://homepages.math.uic.edu/~jan/mcs320s07/Project_Two/sol_body.html

r_1 = np.array([0.97000436, -0.24308753, 0])
v_3 = np.array([-0.93240737, -0.86473146, 0])
v_2 = -v_3 / 2
m = 1

def get_og_figure_8():
    return NBody(np.array([r_1, -r_1, [0, 0, 0]]),
                     np.array([v_2, v_2, v_3]),
                     m * np.ones(shape=(3,)), collision_tolerance=None)



E_0 = -1.2871419917663258

def get_figure_8(v_1, y_1, collision_tolerance = 10**(-4), escape_tolerance = 10**4):
    """
    Produces an NBody instance by perturbing the Figure 8 3-Body configuration whilst conserving energy and angular momentum
    :param v_1: the desired velocity vector of one of the side bodies.
                The other side body will have this same velocity, whilst the central body will have velocity -2*v_1
    :param y_1: the desired y coordinate of one of the side bodies
    :return: an NBody instance for a perturbed Figure 8
    """

    # check: velocity on xy plane
    assert v_1[2] == 0

    # calculate the x-coordinate of position of one of the side bodies given the desired velocity and y-coordinate
    # ensures that energy is conserved
    squared_x_component = (5/(2*(3*m*nm.ten_norm(v_1, axis = 0, sqrt = False) - E_0))) ** 2 - y_1 ** 2
    check_exception(squared_x_component >= 0, Figure8InitException, msg = f"Attempted to root {squared_x_component} when initialising Figure 8")
    x_component = np.sqrt(squared_x_component)

    # compute initial velocities of the 3 bodies
    init_velocities = np.array([v_1, v_1, -2*v_1])

    # compute initial position of a side body
    x_1 = np.array([x_component, y_1, 0])

    # compute initial positions of the 3 bodies
    init_positions = np.array([x_1, -x_1, [0,0,0]])

    nbod = NBody(init_positions, init_velocities, m*np.ones(shape = (3,)), collision_tolerance=collision_tolerance, escape_tolerance=escape_tolerance)

    assert abs(nbod.energy - E_0) < 10**(-6)

    return nbod






import matplotlib.pyplot as plt
from matplotlib import colors
import os
import json

from nbodysim.three_body import *
from nbodysim.exceptions import *
from nbodysim.integrators.leapfrog_3 import Leapfrog3

class StabilityPlotter():
    """
    Superclass, used when calculating and plotting an stability matrix.
    A stability matrix is a matrix, with each cell containing a NBody instance, on which a simulation is run.
    Each NBody instance is a Figure of 8, albeit with perturbations to its x and y components of velocity (v_x and v_y respectively)
    The outcome of the simulation is associated with a number, which is then placed within the cell.
    The resulting matrix can be plotted.
    """
    def __init__(self, perturb, n_trials, collision_tolerance, escape_tolerance, centre_x = 0, centre_y = 0, steps = 10**4, delta = 10**-2, tolerance = 10**-2, adaptive_constant = 0.1, delta_lim = 10**-5):
        """
        :param perturb: the unit amount by which v_x and v_y can be perturbed.
                        That is,any perturbation to vx or vy is a multiple of perturb
        :param n_trials: the number of times that the initial conditions are perturbed to one side of 0.
                         For example, if n_trials = 5, then we will produce a 11 x 11 matrix.
        :param collision_tolerance: maximum distance between 2 bodies allowed before ending simulation
                                    If None, collisions are not considered.
        :param escape_tolerance: maximum distance away from the centre of mass (COM) allowed before ending simulation
                                 If None, escape_tolerance is automatically calculated
                                 If -1, escape_tolerance is not considered
        :param centre_x: the x coordinate of the centre of the stability plot
        :param centre_y: the y coordinate of the centre of the stability plot
        :param steps: the number of steps to integrate for.
                      Since adaptive time step will be used, this is simply used alongside delta to calculate simulation time
        :param delta: time step to use for the integrator. Smaller time steps lead to more accurate orbits.
                      Since adaptive time step will be used, this is simply used alongside steps to calculate simulation time
        :param tolerance: allowed absolute error for determining conservation of calculated quantities.
        :param adaptive_constant: constant resizing factor for adaptive time step
        :param delta_lim: smallest value allowed for the adaptive time step
        """

        self.perturb = perturb
        self.n_trials = n_trials
        self.collision_tolerance = collision_tolerance
        self.escape_tolerance = escape_tolerance
        self.centre_x = centre_x
        self.centre_y = centre_y
        self.steps = steps
        self.delta = delta
        self.run_time = steps * delta
        self.tolerance = tolerance
        self.adaptive_constant = adaptive_constant
        self.delta_lim = delta_lim

        # encode errors as numbers
        self.exception_dict = {"SmallAdaptiveDeltaException": 2,
                              "COMNotConservedException": 3,
                              "BodyEscapeException": 4,
                              "BodyCollisionException": 5,
                              "Figure8InitException": 6,
                              "LinearMomentumNotConservedException": 7,
                              "AngularMomentumNotConservedException": 8,
                              "EnergyNotConservedException": 9}

    def get_error_score(self, exception, time_elapsed):
        """
        Modifies the numerical value of an error (as described in the exception_dict),
        by accounting for the time taken (in simulation time units) until the error appears.
        Thus, a higher error_score indicates that the error took longer to appear, so the orbit was more stable
        :param exception: the exception which caused an error in the simulation
        :param time_elapsed: time until the exception appear
        :return: a numerical value, encompassing the type of error, alongside the time taken until the error happens
        """

        # calculate the factor to account for time taken until error occurred
        addon = time_elapsed/self.run_time

        # to avoid errors being confused, ensure that addon never reaches or exceeds 1
        while addon >= 1:
            addon -= 10**-4

        return self.exception_dict[exception.__class__.__name__] + (1 - addon)

    def get_stability_matrix(self):
        """
        Perturbs Figure 8 orbit, producing a matrix outlining stability regions.
        :return: a (2*n_trials + 1 x 2*n_trials + 1), with numbers representing errors dependent on how the simulation ended.
                 Error-to-number conversion depedns on exception_dict
        """

        # calculate size of matrix
        n = 2 * self.n_trials + 1

        # initialise stability matrix
        stability_matrix = np.zeros(shape=(n, n))

        # the amount by which y component of velocity is changed
        dvy = self.perturb * self.n_trials

        for i in range(n):
            # the amount by which x component of velocity is changed
            dvx = -self.perturb * self.n_trials
            for j in range(n):
                print(f"{i},{j}")
                # initialise perturbed figure of 8
                try:
                    # check for potential exceptions (either Figure of 8 or adaptive delta) during initialisation
                    nbody = get_figure_8(-0.5 * np.array([-0.93240737, -0.86473146, 0]) + np.array([dvx + self.centre_x, dvy + self.centre_y, 0]), -0.24308753,
                                         collision_tolerance=self.collision_tolerance, escape_tolerance=self.escape_tolerance)

                    integrator = Leapfrog3(nbody, steps=self.steps, delta=self.delta, tolerance=self.tolerance, adaptive=True,
                                           adaptive_constant=self.adaptive_constant, store_properties=False, delta_lim=self.delta_lim)

                    # integrate, and catch any exception in the process
                    try:
                        integrator.get_orbits()

                        # if adaptive time step took 10^5 steps (max allowed) set value
                        if integrator.full_run:
                            stability_matrix[i, j] = 1

                    except(SmallAdaptiveDeltaException,
                           COMNotConservedException,
                           LinearMomentumNotConservedException,
                           BodyEscapeException,
                           BodyCollisionException,
                           AngularMomentumNotConservedException,
                           EnergyNotConservedException) as e:

                        # set matrix entry according to error
                        stability_matrix[i, j] = self.get_error_score(e, integrator.times[-1])
                except (Figure8InitException,
                        SmallAdaptiveDeltaException) as e:

                    # set matrix entry according to error
                    stability_matrix[i, j] = self.exception_dict[e.__class__.__name__]

                dvx += self.perturb
            dvy -= self.perturb

        return stability_matrix

    def stability_matrix_to_json(self, stability_matrix: np.ndarray, filename):
        """
        Saves calculated stability matrix, alongside other values used during initialisation of instance,
        as JSON.
        :param stability_matrix: the stability matrix to save
        :param filename: the path to which the stability matrix is saved
        :return: 0 if saving was succesful, 1 otherwise
        """
        json_dict = {"perturb": self.perturb,
                     "n_trials": self.n_trials,
                     "collision_tolerance": self.collision_tolerance,
                     "escape_tolerance": self.escape_tolerance,
                     "centre_x": self.centre_x,
                     "centre_y": self.centre_y,
                     "steps": self.steps,
                     "delta": self.delta,
                     "tolerance": self.tolerance,
                     "adaptive_constant": self.adaptive_constant,
                     "delta_lim": self.delta_lim,
                     "stability_matrix": stability_matrix.tolist()}

        try:
            if not os.path.exists(filename):
                open(filename, "w").close()
                zero_file = True
            else:
                zero_file = os.path.getsize(filename) == 0

            with open(filename, "r+") as json_file:

                if zero_file:
                    json_rewrite = {"stabinv": [json_dict]}
                else:
                    json_rewrite = json.load(json_file)

                    if "stabinv" in json_rewrite:
                        assert isinstance(json_rewrite["stabinv"], list)
                        json_rewrite["stabinv"].append(json_dict)
                    else:
                        json_rewrite = {"stabinv": [json_dict], "errors": [json_rewrite]}

                json_file.seek(0)
                json_file.truncate()

                json.dump(json_rewrite, json_file, indent = 2)

            return 0
        except IOError as e:
            print("Error saving stability params")
            print(e)
            return 1

    def plot_stability_matrix(self, stability_matrix = None, n_ticks = 10, grad = False, show = True, save_fig = False, save_matrix = True, **kwargs):
        """
        Plots an stability matrix given the parameters
        :param stability_matrix: the stability matrix to plot. If None, automatically calculates it.
        :param n_ticks: the number of ticks to use when plotting
        :param grad: if True, plot will use a gradient to represent numbers. Otherwise, will use discrete colours for each number.
        :param show: if True, displays the plot
        :param save_fig: if True, saves the resulting image
        :param save_matrix: if True, saves StabilityPlotter arguments (i.e delta_lim), alongside the produced stability matrix, all in JSON
        :param kwargs: expect at most 2 arguments:
                       json_name: file path to JSON file on which to save StabilityPlotter data, should save_matrix be True
                       fig_name: file path to to save the figure produced, should save_fig be True. Automaticlaly generates file path, if none is provided.
        """

        # calculate stability matrix if not provided
        if stability_matrix is None:
            stability_matrix = self.get_stability_matrix()

        # save stability matrix if required
        if save_matrix:
            if "json_name" in kwargs:
                self.stability_matrix_to_json(stability_matrix, kwargs["json_name"])
            else:
                self.stability_matrix_to_json(stability_matrix, "../img_resources/report_data/report_jsons.json")

        # calculate the side length (to one side of 0) required for the plot
        # for example, if perturb = 0.1 and n_trials = 10,
        # we expect that at most we perturb the initial conditions by 1 in the x and y direction (or negative)
        extent_lim = self.n_trials*self.perturb

        # set up the figure and plot the matrix, with numbers giving pixel values
        fig, ax = plt.subplots()

        if grad:
            # plot using continuous colours
            cmap = plt.get_cmap("nipy_spectral")
            cax = ax.imshow(stability_matrix,
                            extent=(-extent_lim - self.perturb * 0.5 + self.centre_x,
                                    extent_lim + self.perturb * 0.5 + self.centre_x,
                                    -extent_lim - self.perturb * 0.5 + self.centre_y,
                                    extent_lim + self.perturb * 0.5 + self.centre_y),
                            cmap=cmap, vmin=0, vmax=10)
        else:
            # plot using discrete colours
            cmap = plt.get_cmap("viridis")
            # since errors are discrete numbers, use this for the cmap
            norm = colors.BoundaryNorm(np.arange(-0.5, 10, 1), cmap.N)

            cax = ax.imshow(stability_matrix,
                            extent=(-extent_lim - self.perturb * 0.5 + self.centre_x,
                                    extent_lim + self.perturb * 0.5 + self.centre_x,
                                    -extent_lim - self.perturb * 0.5 + self.centre_y,
                                    extent_lim + self.perturb * 0.5 + self.centre_y),
                            norm=norm, cmap=cmap, vmin=0, vmax=9)

        # set plot properties
        ax.set_xlabel(r"$\Delta v_x$")
        ax.set_ylabel(r"$\Delta v_y$")
        ax.set_title("Stability of Figure 8 Under Perturbations", pad=20)

        ax.set_xticks(np.arange(-extent_lim + self.centre_x, extent_lim + self.perturb + self.centre_x, step=2 * extent_lim / n_ticks))
        ax.set_yticks(np.arange(-extent_lim + self.centre_y, extent_lim + self.perturb + self.centre_y, step=2 * extent_lim / n_ticks))

        # show a colorbar
        cb = plt.colorbar(cax, ticks = range(0,10))
        cb.set_label("Error Type", labelpad=20)

        plt.tight_layout()

        # save the plot
        if save_fig and "fig_name" in kwargs:
            if kwargs["fig_name"] is None:
                int_to_string = lambda x: str(x).replace(".", "_")
                save_string = f"stability{2*self.n_trials + 1}-perturb{int_to_string(self.perturb)}-time{int(self.steps*self.delta)}-AC{int_to_string(self.adaptive_constant)}-DL{int_to_string(self.delta_lim)}-ET{int_to_string(self.escape_tolerance)}-CT{int_to_string(self.collision_tolerance)}-TOL{int_to_string(self.tolerance)}-({self.centre_x}-{self.centre_y})"
            else:
                save_string = kwargs["fig_name"]

            # ensure that we don't overwrite an already existing image
            idx = 2
            filename = save_string
            save_file = filename + ".png"
            # if we are to overwrite a file, check & modify the new file accordingly
            while os.path.exists(save_file):
                filename = save_string + f"({idx})"
                save_file = filename + ".png"
                idx +=1

            plt.savefig(save_file)

        if show:
            plt.show()




import json
from multiprocessing import Pool
import numpy as np
import os
import tqdm
from matplotlib import pyplot as plt

from nbodysim.integrators.leapfrog_3 import Leapfrog3
from nbodysim.stability_investigator.stability_plotter import StabilityPlotter
from nbodysim.stability_investigator.mp_stability_plotter import MPStabilityPlotter
from nbodysim.three_body import get_figure_8
from nbodysim import nmath as nm

class StabilityAnalyser():
    """
    Class used to more thoroughly analyse the regions of stability calculated by a StabilityPlotter instance.
    Requires that the stability matrix be calculated or store somewhere.
    It is NOT responsible of calculating it in the first place.
    In particular, considers how any set of stable initial conditions compare to the original Figure of 8,
    assigning a stabiliity score which can then be visualised as a stability matrix
    """

    def __init__(self, stability_plotter: StabilityPlotter, stability_matrix, **kwargs):
        """
        :param stability_plotter: the stability plotter instance to analyse.
                                  If None, requires kwargs.
        :param stability_matrix: the stabiliy matrix to analyse.
                                 If None, requires kwargs.
        :param kwargs: contains arguments, should a stability_plotter or stability_matrix be provided as None
                       - should contain all initialisation arguments of a StabilityPlotter. That is:
                            -> perturb
                            -> n_trials
                            -> collision_tolerance
                            -> escape_tolerance
                            -> centre_x
                            -> centre_y
                            -> steps
                            -> delta
                            -> tolerance
                            -> adaptive_constant
                            -> delta_lim
                        - additionally, if no stability_matrix is provided, should contain information to obtain it from a JSON file
                            -> filename: file path to JSON
                            -> idx: if None, returns all stability matrices found in filename; otherwise, returns the (idx - 1)th matrix found
                            -> use_self: alternatively to idx, use the StabilityPlotter instance to search through the JSON file
                                        & find the stability matrix that would be produced by the StabilityPlotter instance
        """

        self.stability_plotter = stability_plotter

        # if no stability_plotter is provided, use kwargs to initialise
        if stability_plotter is None:
            try:
                self.perturb = kwargs["perturb"]
                self.n_trials = kwargs["n_trials"]
                self.collision_tolerance = kwargs["collision_tolerance"]
                self.escape_tolerance = kwargs["escape_tolerance"]
                self.centre_x = kwargs["centre_x"]
                self.centre_y = kwargs["centre_y"]
                self.steps = kwargs["steps"]
                self.delta = kwargs["delta"]
                self.run_time = self.steps * self.delta
                self.tolerance = kwargs["tolerance"]
                self.adaptive_constant = kwargs["adaptive_constant"]
                self.delta_lim = kwargs["delta_lim"]
                self.stability_plotter = MPStabilityPlotter(perturb=self.perturb, n_trials=self.n_trials, collision_tolerance = self.collision_tolerance, escape_tolerance = self.escape_tolerance, steps = self.steps, delta = self.delta, tolerance = self.tolerance, adaptive_constant = self.adaptive_constant, delta_lim = self.delta_lim)
            except KeyError as e:
                print("StabilityAnalyser initialisation failed, due to a key parameter not being provided.")
                print("Required parameters: perturb, n_trials, collision_tolerance, escape_tolerance, steps,\n"
                      "delta, tolerance, adaptive_constant, delta_lim")

                raise e
        else:
            self.perturb = stability_plotter.perturb
            self.n_trials = stability_plotter.n_trials
            self.collision_tolerance = stability_plotter.collision_tolerance
            self.escape_tolerance = stability_plotter.escape_tolerance
            self.centre_x = stability_plotter.centre_x
            self.centre_y = stability_plotter.centre_y
            self.steps = stability_plotter.steps
            self.delta = stability_plotter.delta
            self.run_time = self.steps * self.delta
            self.tolerance = stability_plotter.tolerance
            self.adaptive_constant = stability_plotter.adaptive_constant
            self.delta_lim = stability_plotter.delta_lim

        if stability_matrix is None:
            try:
                self.stability_matrix = self.stability_matrix_from_json(kwargs["filename"], kwargs["idx"], kwargs["use_self"])
            except KeyError as e:
                print("StabilityAnalyser initialisation failed, due to a key parameter not being provided.")
                print("Required parameters: filename, idx and use_self")

                raise e
        else:
            self.stability_matrix = stability_matrix

        self.updated_stability_matrix = False


    def stability_matrix_from_json(self, filename, idx = None, use_self = False):
        """
        Reads stability_matrix from JSON.
        Should be used to read data from a JSON file created by stability_matrix_to_json in the StabilityPlotter class.
        :param filename: file path to JSON
        :param idx: if None, returns all stability matrices found in filename; otherwise, returns the (idx - 1)th matrix found
        :param use_self: alternatively to idx, use the StabilityPlotter instance to search through the JSON file
                                        & find the stability matrix that would be produced by the StabilityPlotter instance
        :return: if idx is None, a list of stability matrices; otherwise, a stability matrix
        """

        try:
            with open(filename, "r") as json_file:

                # get all the data in the JSON
                sm_json = json.load(json_file)

                # ensure that JSON is in correct format
                if "stabinv" in sm_json:
                    assert isinstance(sm_json["stabinv"], list)
                    sm_dict_list = sm_json["stabinv"]
                    if idx is None:
                        if use_self:

                            self_dict = {"perturb": self.perturb,
                                         "n_trials": self.n_trials,
                                         "collision_tolerance": self.collision_tolerance,
                                         "escape_tolerance": self.escape_tolerance,
                                         "centre_x": self.centre_x,
                                         "centre_y": self.centre_y,
                                         "steps": self.steps,
                                         "delta": self.delta,
                                         "tolerance": self.tolerance,
                                         "adaptive_constant": self.adaptive_constant,
                                         "delta_lim": self.delta_lim}

                            for sm_dict in sm_dict_list:
                                # go through each stability matrix,
                                # and if found, returns the stability matrix with parameters as set during init
                                stability_matrix = np.array(sm_dict.pop("stability_matrix"))
                                common_dict = dict(sm_dict.items() & self_dict.items())
                                if len(common_dict) == 9 or len(common_dict) == 11:
                                    return stability_matrix

                            print(f"No stability_matrix was found in {filename} with parameters {self_dict}")
                            return np.zeros(shape=(1,))
                        else:
                            # if no idx is provided, return all the stability matrices found
                            return np.array([dict["stability_matrix"] for dict in sm_dict_list])
                    else:
                        # ensure that idx is within bounds and return the stability matrix found
                        assert len(sm_dict_list) > idx >= 0, f"Index {idx} out of bounds for dictionary list of length {len(sm_dict_list)} "
                        return np.array((sm_dict_list[idx])["stability_matrix"])
                else:
                    print(f"File {filename} in incorrect format. Should have a single key 'stabinv' with value as a list.")
                    return np.zeros(shape=(1,))

        except IOError as e:
            print(f"An error occurred when attempting to open {filename} ")
            print(e)
            return np.zeros(shape=(1,))

    def get_positions_cell(self, coords):
        """
        Similar to MPStabilityPlotter method get_stability_cell.
        Computes the orbit for a single cell in the stability matrix (expecing the coords to correspond to stable regions in the matrix=
        :param coords:  the row and column of the cell (with coords = (row, col)). Used to calculate the perturbations.
        :return: a dictionary, containing the row and column that lead to the NBody instance,
                 alongside the positions of each body during the orbit.
        """

        # extract row and column, in order to calculate perturbation
        row, col = coords

        # the maximum allowed perturbation
        n = self.perturb * self.n_trials

        # the amount by which y component of velocity is changed
        dvy = n - row * self.perturb

        # the amount by which x component of velocity is changed
        dvx = -n + col * self.perturb

        nbody = get_figure_8(-0.5 * np.array([-0.93240737, -0.86473146, 0]) + np.array([dvx + self.centre_x, dvy + self.centre_y, 0]), -0.24308753,
                                         collision_tolerance=self.collision_tolerance, escape_tolerance=self.escape_tolerance)

        integrator = Leapfrog3(nbody, steps=self.steps, delta=self.delta, tolerance=self.tolerance, adaptive=True,
                               adaptive_constant=self.adaptive_constant, store_properties=False, delta_lim=self.delta_lim)

        integrator.get_orbits()

        return {"row": row,
                "col": col,
                "positions": integrator.position_orbit}

    def get_stable_positions(self):
        """
        Similar to MPStabilityPlotter method get_stability_matrix.
        Using the stability_matrix computed during initialisation, applies get_position_cell for each element of the stability region.
        Uses multiprocessing to compute these values.
        :param coords:  the row and column of the cell (with coords = (row, col)). Used to calculate the perturbations.
        :return: a list of dictionaries, eahc containing the row and column that lead to the NBody instance,
                 alongside the positions of each body during the orbit.
        """

        # list containing the arguments to be used when creating the dictionaries
        args = []

        # size of the stability_matrix
        n = 2 * self.n_trials + 1

        for i in range(n):
            for j in range(n):
                # get the coords if and only if the region is stable
                if self.stability_matrix[i,j] == 0 or self.stability_matrix[i,j] == 1:
                    args.append((i, j))

        # calculate number of CPU cores that can be used
        n_cpus = os.cpu_count()

        # use pooling to multiprocess across a list of different arguments
        # use tqdm in order to get progress in calculations
        with Pool(n_cpus) as pool:
            results = list(tqdm.tqdm(pool.imap(func=self.get_positions_cell, iterable=args), total=len(args)))

        return results

    def get_stability_scores(self, square_size):
        """
        Computes the stability score for each stable orbit.
        This is done by comparing the stable orbit to the original Figure of 8.
        To do this, we change the stable orbit positions to a new set of coordinates,
        such that the 3 body positions can be represented by 2 coordinates, X_1 and X_2.
        If we plot X_1 and X_2 in a grid of squares, we can the count the number of distinct squares that the curve goes through.
        The number of squares is then the stability score.
        :param square_size: the length of a square in the grid
        :return: a list of dictionaries, containing the row, col and positions of the stable NBody instance been experimented,
        alongside the transformed coordinates, X_1 and X_2, and the stability score
        """

        # compute the positions of each of the stable orbits
        stability_positions = self.get_stable_positions()
        stability_scores = []

        # for each stable position, compute the stability score
        for stable_position in stability_positions:

            # use a dictionary to keep track of the squares visited
            visited_squares = {}

            # use the positions to transform to X_1, X_2 coords
            X_1, X_2 = nm.get_relative_normalised_positions(stable_position["positions"])

            # compute the the index of the x coordinate of the square containing X_1
            x = np.floor(X_1/square_size)
            # compute the the index of the y coordinate of the square containing X_1
            y = np.floor(X_2 / square_size)

            # compute a unique number based on x and y
            keys = nm.pair_encoder(x, y)

            # check if a square has been visited before
            for key in keys:
                if key not in visited_squares:
                    visited_squares[key] = 1

            position_dict = stable_position
            position_dict["X_1"] = X_1
            position_dict["X_2"] = X_2
            position_dict["stability_score"] = len(visited_squares)

            stability_scores.append(position_dict)

        return stability_scores

    def update_stability_matrix(self, sb_scores, square_size = 0.01):
        """
        Using a list of dictionaries (containing stability scores), updates the stability matrix.
        Stored as an attribute u_stability_matrix
        :param sb_scores: the stability scores. If None, computes it automatically.
        :param square_size: square_size to be used if sb_scores is None
        """

        # compute stability_scores if not provided
        if sb_scores is None:
            sb_scores = self.get_stability_scores(square_size)

        # intialise updated stability_matrix
        self.u_stability_matrix = self.stability_matrix
        self.updated_stability_matrix = True

        # use the largest stability score to normalise all stability scores between 0 and 1
        normalisation_factor = max([sb_score["stability_score"] for sb_score in sb_scores])
        normalisation_factor = normalisation_factor + 10**(np.floor(np.log10(normalisation_factor)) - 1)

        # for each stability score dictionary, update the stability matrix
        for sb_score in sb_scores:
            stability_value = self.stability_matrix[sb_score["row"], sb_score["col"]] + sb_score["stability_score"]/normalisation_factor
            self.u_stability_matrix[sb_score["row"], sb_score["col"]] = stability_value

    def plot_stability_score(self, sb_score, square_size = 0.01, idx = 0):
        """
        Using a stability score dictionary,
        plots the perturbed Figure of 8 orbit, alongside the orbit in the X_1 and X_2 coordinates
        :param sb_score: the stability score dictionary to use. If None, computes it.
        :param square_size: the square size to use to compute the stability scores
        :param idx: the index of the stability score dictionary of interest within the stability scores list
        """

        # obtain the stability score of interest
        if sb_score is None:
            sb_score: dict = self.get_stability_scores(square_size)[idx]

        # plot the Figure of 8 orbit alongside the X_1,X_2 orbit
        fig, ax = plt.subplots(1, 2, figsize=(10, 5))
        ax[0].plot(sb_score["positions"][0, :, 0], sb_score["positions"][0, :, 1], c="k")
        ax[0].plot(sb_score["positions"][1, :, 0], sb_score["positions"][1, :, 1], c="k")
        ax[0].plot(sb_score["positions"][2, :, 0], sb_score["positions"][2, :, 1], c="k")
        ax[0].set_title("Figure of 8")
        ax[1].plot(sb_score["X_1"], sb_score["X_2"])
        ax[1].set_title(f"Shifted Coords Figure of 8\nScore: {sb_score['stability_score']}")
        plt.show()

    def plot_stability_scores(self, sb_scores, square_size = 0.01, **kwargs):
        """
        Using a list of stability score dictionaries,
        plots the perturbed Figure of 8 orbit, alongside the orbit in the X_1 and X_2 coordinates.
        Can be used for a list of stability score dictionaries, or a single stability score dictionary
        :param sb_scores: the stability scores dictionary to use. If None, computes it.
        :param square_size: the square size to use to compute the stability scores
        :param kwargs: should contain a single argument:
                        ->idx: the index of the stability score dictionary of interest within the stability scores list
        """

        # compute list of stability scores
        if sb_scores is None:
            sb_scores = self.get_stability_scores(square_size)

        # if idx provided, we want a single stability score dictionary, so get it and plot it
        if "idx" in kwargs:
            sb_score = sb_scores[kwargs["idx"]]
            self.plot_stability_score(sb_score)
        else:
            # otherwise, plot all stability score dictionaries
            for sb_score in sb_scores:
                self.plot_stability_score(sb_score)

    def plot_updated_stability_matrix(self, n_ticks = 10, grad = False, show = True, save_fig = False, save_matrix = True, **kwargs):
        """
        Plots an updated stability matrix, containing information on the stability of the stable regions.
        If the stability matrix wasn't updated, providing a list of stability score dictionaries can be used to update it.
        :param n_ticks: the number of ticks to use when plotting
        :param grad: if True, plot will use a gradient to represent numbers. Otherwise, will use discrete colours for each number.
        :param show: if True, displays the plot
        :param save_fig: if True, saves the resulting image
        :param save_matrix: if True, saves StabilityPlotter arguments (i.e delta_lim), alongside the produced stability matrix, all in JSON
        :param kwargs: expect at most 4 arguments:
                       sb_scores: list of dictionaries containing the stability scores. If None, will be automatically computed.
                       square_size: if sb_scores is None, used to compute them
                       json_name: file path to JSON file on which to save StabilityPlotter data, should save_matrix be True
                       fig_name: file path to to save the figure produced, should save_fig be True. Automaticlaly generates file path, if none is provided.
        """

        if not self.updated_stability_matrix:
            if "sb_scores" in kwargs and "square_size" in kwargs:
                self.update_stability_matrix(kwargs["sb_scores"], kwargs["square_size"])
            else:
                print("sb_scores or square_size not provided as arguments, and stability matrix wasn't updated. Plotting original stability matrix.")
                self.stability_plotter.plot_stability_matrix(stability_matrix = self.stability_matrix, n_ticks=n_ticks, grad = grad, show = show, save_fig = save_fig, save_matrix = save_matrix, **kwargs)

        self.stability_plotter.plot_stability_matrix(stability_matrix = self.u_stability_matrix, n_ticks=n_ticks, grad = grad, show = show, save_fig = save_fig, save_matrix = save_matrix, **kwargs)








from nbodysim.integrator import Integrator
from nbodysim import nmath as nm

class EulerCromer(Integrator):
    """
    Class defining an integrator via the Euler-Cromer Method
    """
    def __init__(self, nbody, steps, delta, tolerance = 1e-6, adaptive = False, adaptive_constant = 1, delta_lim =10 ** -5, store_properties = False):
        """
        :param nbody: NBody instance which we integrate
        :param steps: the number of steps to integrate for
        :param delta: timestep to use for the integrator. Smaller timesteps lead to more accurate orbits.
        :param tolerance: allowed absolute error for determining conservation of calculated quantities
        :param adaptive: if True, the Integrator will use an adaptive timestep (instead of a fixed one)
        :param adaptive_constant: constant used when calculating adaptive timestep. Smaller adaptive_constant leads to more accurate orbits.
        """

        # execute initialisation from superclass
        super().__init__(nbody, steps, delta, tolerance = tolerance, adaptive = adaptive, adaptive_constant= adaptive_constant, delta_lim = delta_lim, store_properties = store_properties)

    def integration_step(self, t, delta):
        """
        Integration step for the Euler-Cromer method.
        v_t = v_{t-1} + a_{t-1}*Δt
        x_t = x_{t-1} + v_{t}*Δt
        :param t: the step at which the calculation is made
        """

        acc_t = self.nbody.get_acceleration()

        new_velocities = self.velocity_orbit[:, t-1, :] + delta * acc_t
        new_positions = self.position_orbit[:, t-1, :] + delta * new_velocities

        return new_positions, new_velocities, None



import math
import random

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation
from matplotlib.gridspec import GridSpec
from mpl_toolkits import mplot3d

from nbodysim import nmath as nm

from matplotlib.axes._axes import _log as matplotlib_axes_logger
matplotlib_axes_logger.setLevel('ERROR')

class OrbitPlotter:
    """
    Class used to plot the results of using an integrator
    """
    def __init__(self, integrator, seed = 42, animation_steps = 1, twodims = True, grid = False):
        """
        :param integrator: Integrator instance which has calculated orbits
        :param seed: used as seed to generate the colours of the plot
        :param animation_steps: number of steps to consider per frame of animation
        :param twodims: if True, will plot the orbit using 2 dimensions (x and y)
        :param grid: if True, will plot in 2 dimensions,
                     whilst displaying information, such as energy, angular momentum, position and velocity change across time
        """

        self.integrator = integrator
        self.seed = seed
        self.animation_steps = animation_steps
        self.twodims = twodims
        self.grid = grid

        # take properties of Integrator (used in calculations when plotting)
        self.n = integrator.nbody.n
        self.position_orbit = integrator.position_orbit
        self.delta = integrator.delta
        self.steps = integrator.steps

        # generate colours for the orbits
        self.colours = self.generate_colour()

        # calculate all the times at which the orbit positions/velocities were calculated
        # if adaptive, times are calculated within the integration
        self.times = self.integrator.times if self.integrator.adaptive else np.array([i*self.delta for i in range(self.steps)])

        # set up the figure for plotting
        self.set_fig()

    def generate_colour(self):
        """
        Uses the seed from init to generate a random list of n colours, one for each orbit
        """
        random.seed(self.seed)
        colours = []

        for i in range(self.n):
            r = random.random()
            g = random.random()
            b = random.random()
            colours.append((r, g, b))

        return colours

    def set_fig(self):
        """
        Sets up the fig on which the plots are drawn.
        If grid = True, we use GridSpec to be able to display not only the orbits,
        but values such as positions, velocities, energies and angular momentum.
        Otherwise, if twodims = True, we set a 2D plot, and if twodims = False, we set a 3D plot.
        """

        if not self.grid:
            if self.twodims:
                self.fig, self.ax = plt.subplots()
            else:
                self.fig = plt.figure()
                self.ax = plt.axes(projection="3d")
        else:
            self.fig = plt.figure(figsize = (14, 8))
            self.gs = GridSpec(nrows = 3, ncols = 5, figure = self.fig)

        # set window title
        self.fig.canvas.set_window_title("Vacation Scholarship - N-Body Problem")

# ---------------------------------------- ANIMATED PLOTS ----------------------------------------

    def animation_func_orbit_3D(self, frame):
        """
        Function for drawing a frame in a 3D animation
        :param frame: the frame in animation to draw on
        """

        # clear the plot
        plt.cla()

        # for each body, plot frame * animation_steps steps in the orbit
        for i,orbit in enumerate(self.position_orbit[:,:frame*self.animation_steps,:]):
            # display the legend only for the last frame
            if frame == self.n - 1:
                self.ax.plot3D(orbit[:,0], orbit[:,1], orbit[:,2], label = f"Orbit {i+1}", c = self.colours[i])
                self.ax.legend()
            else:
                self.ax.plot3D(orbit[:,0], orbit[:,1], orbit[:,2], c = self.colours[i])

            # display the position of the body at the end of the plotted interval
            # gives an idea of the direction in which the bodies are moving
            if (len(orbit) > 0):
                self.ax.scatter3D(orbit[-1, 0], orbit[-1, 1], orbit[-1,2], c = self.colours[i], s=10)

        # display time of the simulation only for constant timestep
        if not self.integrator.adaptive:
            self.ax.annotate(f"t = {round(frame * self.delta * self.animation_steps, 1)}", xy=self.integrator.nbody.com[:2])

    def animation_func_orbit_2D(self, frame):
        """
        Function for drawing a frame in a 2D animation
        :param frame: the frame in animation to draw on
        """

        # clear the plot
        plt.cla()

        # for each body, plot frame * animation_steps steps in the orbit
        for i,orbit in enumerate(self.position_orbit[:,:np.int(np.floor(frame*self.animation_steps)),:]):
            # display the legend only for the last frame
            if frame == self.n - 1:
                self.ax.plot(orbit[:,0], orbit[:,1], label = f"Orbit {i+1}", c = self.colours[i])
                self.ax.legend()
            else:
                self.ax.plot(orbit[:,0], orbit[:,1], c = self.colours[i])

            # display the position of the body at the end of the plotted interval
            # gives an idea of the direction in which the bodies are moving
            if (len(orbit) > 0):
                self.ax.scatter(orbit[-1, 0], orbit[-1, 1], c = self.colours[i], s=10)

        # create title for plot
        self.ax.set_title(f"Trajectories for a {self.n}-body Problem")

        # display time of the simulation only for constant timestep
        if not self.integrator.adaptive:
            self.ax.annotate(f"t = {round(frame * self.delta * self.animation_steps, 1)}", xy=self.integrator.nbody.com[:2])

    def animate_orbit(self, xlims = (-5,5), ylims = (-5,5)):
        """
        Creates and shows the animated orbits from the simulations
        :param xlims: tuple giving the x limit for the plotting
        :param ylims: tuple giving the y limit for the plotting
        """

        assert (not self.grid)

        if self.twodims:
            # create 2D animation
            animator = FuncAnimation(self.fig, self.animation_func_orbit_2D,
                                     frames = math.ceil(self.steps/self.animation_steps), repeat = False, interval = 100)

            # x and y axes are same size
            plt.axis("equal")
        else:
            # create 3D animation
            animator = FuncAnimation(self.fig, self.animation_func_orbit_3D,
                                     frames = math.ceil(self.steps/self.animation_steps), repeat = False, interval = 100)

            # label the z axis (since we are considering 3D plotting)
            self.ax.set_zlabel("z")

        # set plot properties
        self.ax.set_xlim(xlims)
        self.ax.set_ylim(ylims)
        self.ax.set_xlabel("x")
        self.ax.set_ylabel("y")

        # display animation
        plt.show()

# ---------------------------------------- STATIONARY PLOT ----------------------------------------

    def plot_orbit(self, xlims = (-5,5), ylims = (-5,5), size = 50):
        """
        Plots the orbit calculated from the integrator as a standard plot
        :param xlims: tuple giving the x limit for the plotting
        :param ylims: tuple giving the y limit for the plotting
        :param size: size of the scatter marks in the plot
        """

        assert (not self.grid)

        if self.twodims:
            # plot in 2D each orbit, alongisde the start and end points
            for i, orbit in enumerate(self.position_orbit):
                self.ax.plot(orbit[:, 0], orbit[:, 1], label=f"Orbit {i + 1}", c=self.colours[i])
                self.ax.scatter(orbit[0, 0], orbit[0, 1], c=self.colours[i], s = size, marker="^")
                self.ax.scatter(orbit[-1, 0], orbit[-1, 1], c=self.colours[i], s = size)
                self.ax.set_title(f"Trajectories for a {self.n}-body Problem")
                # plot the orbits produced by the simulation
        
            
        else:
            # plot in 3D each orbit, alongisde the start and end points
            for i,orbit in enumerate(self.position_orbit):
                self.ax.plot3D(orbit[:,0], orbit[:,1], orbit[:,2], label = f"Orbit {i+1}", c = self.colours[i])
                self.ax.scatter3D(orbit[0, 0], orbit[0, 1], orbit[0, 2], c = self.colours[i], s = size, marker = "^")
                self.ax.scatter3D(orbit[-1, 0], orbit[-1, 1], orbit[-1, 2], c = self.colours[i], s = size)
                self.ax.set_title(f"Trajectories for a {self.n}-body Problem", pad=40)
                self.ax.set_zlabel("z")

        # set plot properties
        self.ax.set_xlim(xlims)
        self.ax.set_ylim(ylims)
        self.ax.set_xlabel("x")
        self.ax.set_ylabel("y")
        self.ax.axis("equal")
        self.ax.legend()

        # display plot
        plt.show()

# ---------------------------------------- GRID PLOTS ----------------------------------------

    def plot_energies(self, ax, absolute = False, e_components = False):
        """
        Used to display energies across time when plotting in grid mode
        :param ax: the axis object on which the energies are plotted
        :param absolute: if True, displays the values of energy across time.
                         if False, displays the percentage error in the value of energy across time
        :param e_components: if True, shows not only total energy, but also kinetic and GPE
        """

        if absolute:
            # plot energy across time
            ax.plot(self.times, self.integrator.historic_energy, label = "Total Energy", c = "k")

            # if required, plot kinetic and gravitational potential energies
            if e_components:
                ax.plot(self.times, self.integrator.historic_kinetic_energy, label="Kinetic Energy", c = "r")
                ax.plot(self.times, self.integrator.historic_gpe, label="Gravitational Potential Energy", c = "b")

            # set plot properties
            ax.set_title("Energy Change")
            ax.set_xlabel("Time")
            ax.set_ylabel("Energy")
            ax.legend()
        else:
            # plot percentage error in energy across time
            ax.plot(self.times,
                    nm.perc_change(self.integrator.historic_energy[0], self.integrator.historic_energy, perc = True),
                    c = "k")

            # set plot properties
            ax.set_title("Percentage Energy Error")
            ax.set_xlabel("Time")
            ax.set_ylabel("Percentage Error")

    def plot_angular_momentum(self, ax, absolute = False, dim = 2):
        """
        NOTE: THIS METHOD WILL CHANGE WHEN MORE COMPLEX ORBITS ARE CALCULATED,
        AS ANGULAR MOMENTUM WILL NOT ONLY OCCUR IN A SINGLE DIMENSION
        Used to display angular momentum across time when plotting in grid mode
        :param ax: the axis object on which the angular momentum is plotted
        :param absolute: if True, displays the value of angular momentum across time.
                         if False, displays the percentage error in the value of angular momentum across time
        :param dim: the component (x, y or z) of angular momentum to consider
        """

        # get the component of angular momentum according to dim
        # for orbits over the x and y plane, this will be the z direction
        amomentum_z = self.integrator.historic_angular_momentum[:,dim]

        # plot angular momentum either across time, or as percentage error over time
        if absolute:
            ax.plot(self.times, amomentum_z, c="k")
            ax.set_title("Angular Momentum Change")
            ax.set_ylabel("Angular Momentum")
        else:
            ax.plot(self.times, nm.perc_change(amomentum_z[0], amomentum_z, perc=True, init_val=100), c="k")
            ax.set_title("Percentage Angular Momentum Error")
            ax.set_ylabel("Percentage Error")

        ax.set_xlabel("Time")


    def plot_adaptive_delta(self, ax):
        """
        Used to display adaptive delta across time when plotting in grid mode
        :param ax: the axis object on which the angular momentum is plotted
        """

        ax.plot(self.times, self.integrator.historic_delta)
        ax.set_title("Adaptive Delta Change")
        ax.set_xlabel("Time")
        ax.set_ylabel(r"$\Delta t$")

    def plot_dim(self, data, dim, ax, title, legend = False):
        """
        Used to display raw data from a tensor, across one of its dimensions.
        For example, to plot the x position of all the orbits across the simulation,
        as data we would use the position_orbit tensor and dim = 0
        :param data: the tensor containing data to be plotted
        :param dim: the dimension of the tensor to plot
        :param ax: the axis object on which the angular momentum is plotted
        :param title: title for the plot
        :param legend: if True, displays a legend for the plot
        :return:
        """

        # check: the plotting dimension is within the size of the coordinate system
        assert dim < data.shape[-1]

        # extract the data for plotting, according to the desired dimension
        dim_data = data[:,:,dim]

        # ensure tha plotting data is of the expected shape
        assert dim_data.shape == (self.n, self.steps), f"Expected data shape: {(self.n, self.steps)}\n" \
                                                       f"Actual data shape: {dim_data.shape}"

        # for each body in the simulation, plot the approriate data
        for i in range(len(dim_data)):
            ax.plot(self.times, dim_data[i], label = f"Orbit {i + 1}", c = self.colours[i])

        # set plot properties
        ax.set_title(f"{title} Change")
        ax.set_xlabel("Time")
        ax.set_ylabel(title)

        # display legend if required
        if legend:
            ax.legend(loc = "upper right")

    def plot_grid(self, xlims = (-5,5), ylims = (-5,5)):
        """
        Plots the simulation results as a grid containing the resulting orbits, changes in x, y, z position and velocity,
        alongside change in energy and angular momentum (and adaptive delta)
        :param xlims: tuple giving the x limit for the plotting
        :param ylims: tuple giving the y limit for the plotting
        """

        assert self.grid, "To display a grid with the orbit information, initialise OrbitPlotter with grid = True"

        # generate axes objects to plot on
        ax_orbit = self.fig.add_subplot(self.gs[0:2, 0:2])

        ax_position_x = self.fig.add_subplot(self.gs[0,2])
        ax_position_y = self.fig.add_subplot(self.gs[0, 3])
        ax_position_z = self.fig.add_subplot(self.gs[0, 4])

        ax_velocity_x = self.fig.add_subplot(self.gs[1, 2])
        ax_velocity_y = self.fig.add_subplot(self.gs[1, 3])
        ax_velocity_z = self.fig.add_subplot(self.gs[1, 4])

        ax_energy = self.fig.add_subplot(self.gs[2, 0:2])
        ax_amomentum = self.fig.add_subplot(self.gs[2, 3:])

        # plot the orbits produced by the simulation
        for i, orbit in enumerate(self.position_orbit):
            ax_orbit.plot(orbit[:, 0], orbit[:, 1], label=f"Orbit {i + 1}", c = self.colours[i])
            ax_orbit.scatter(orbit[0, 0], orbit[0, 1], c = self.colours[i], s=10, marker="^")
            ax_orbit.scatter(orbit[-1, 0], orbit[-1, 1], c = self.colours[i], s = 10)#*self.integrator.nbody.masses[i])
        ax_orbit.set_title(f"Trajectories for a {self.n}-body Problem")
        ax_orbit.axis("equal")
        ax_orbit.legend()

        # plot energy and angular momentum over time
        self.plot_energies(ax_energy, absolute = False, e_components = True)
        self.plot_angular_momentum(ax_amomentum, absolute = False, dim = 2)

        # plot positions over time
        self.plot_dim(self.position_orbit, dim = 0, ax = ax_position_x, title = r"Position $x$")
        self.plot_dim(self.position_orbit, dim = 1, ax = ax_position_y, title = r"Position $y$")
        self.plot_dim(self.position_orbit, dim = 2, ax = ax_position_z, title = r"Position $z$")

        # plot velocities over time
        self.plot_dim(self.integrator.velocity_orbit, dim = 0, ax=ax_velocity_x, title = r"Velocity $x$")
        self.plot_dim(self.integrator.velocity_orbit, dim = 1, ax = ax_velocity_y, title = r"Velocity $y$")
        self.plot_dim(self.integrator.velocity_orbit, dim = 2, ax = ax_velocity_z, title = r"Velocity $z$")

        # if required, plot the adaptive delta
        if self.integrator.adaptive:
            ax_delta = self.fig.add_subplot(self.gs[2,2])
            self.plot_adaptive_delta(ax_delta)

        # display the grid plots
        plt.tight_layout()
        plt.show()



import numpy as np

from nbodysim.integrators.integrator import Integrator
from nbodysim import nmath as nm

class Leapfrog3(Integrator):
    """
    Class defining an integrator via the 3-Step Leapfrog 2-Step Method
    """
    def __init__(self, nbody, steps, delta, tolerance = 1e-6, adaptive = False, adaptive_constant = 1, delta_lim =10 ** -5, store_properties = False):
        """
        :param nbody: NBody instance which we integrate
        :param steps: the number of steps to integrate for
        :param delta: timestep to use for the integrator. Smaller timesteps lead to more accurate orbits.
        :param tolerance: allowed absolute error for determining conservation of calculated quantities
        :param adaptive: if True, the Integrator will use an adaptive timestep (instead of a fixed one)
        :param adaptive_constant: constant used when calculating adaptive timestep. Smaller adaptive_constant leads to more accurate orbits.
        """

        # execute initialisation from superclass
        super().__init__(nbody, steps, delta, tolerance = tolerance, adaptive = adaptive, adaptive_constant= adaptive_constant, delta_lim = delta_lim, store_properties = store_properties)

        # save acceleration for next iteration.
        # Only require 1 expensive acceleration calculation per step
        self.acc_t = self.nbody.get_acceleration()

    def integration_step(self, t, delta):
        """
        Integration step for the Integer 3-Step Leapfrog method.
        v_{t + 1/2} = v_t + 0.5*a_t*Δt
        x_{t + 1} = x_t + v_{t + 1/2}*Δt
        v_{t + 1} = v_{t + 1/2} + 0.5*a_{t + 1}*Δt
        :param t: the step at which the calculation is made
        :return the new positions and velocities, alongside the acceleration for step t+1
        """

        new_half_velocities = self.velocity_orbit[:, t - 1, :] + self.acc_t * delta * 0.5
        new_positions = self.position_orbit[:, t - 1, :] + new_half_velocities * delta
        acc_tt = self.nbody.get_acceleration(positions=new_positions)
        new_velocities = new_half_velocities + acc_tt * delta * 0.5

        return new_positions, new_velocities, acc_tt


import numpy as np

from nbodysim import nmath as nm
from nbodysim.orbit_plotter import OrbitPlotter

class Integrator:
    """
    Class defining a general integrator, acting as a "superclass" for Euler, Euler-Cromer and all Leapfrog methods
    """
    def __init__(self, nbody, steps, delta, tolerance = 1e-6, adaptive = False, adaptive_constant = 1, delta_lim =10 ** -5, store_properties = False):
        """
        :param nbody: NBody instance which we integrate
        :param steps: the number of steps to integrate for
        :param delta: time step to use for the integrator. Smaller time steps lead to more accurate orbits.
        :param tolerance: allowed absolute error for determining conservation of calculated quantities
        :param adaptive: if True, the Integrator will use an adaptive timestep (instead of a fixed one)
        :param adaptive_constant: constant used when calculating adaptive timestep. Smaller adaptive_constant leads to more accurate orbits.
        """

        self.nbody = nbody

        self.steps = int(steps)

        self.adaptive = adaptive
        self.adaptive_constant = adaptive_constant
        self.delta_lim = delta_lim

        self.tolerance = tolerance

        self.delta = delta

        self.store_properties = store_properties

        # creates all arrays used by integrator
        # these hold the positions and velocities of the calculated orbits,
        # alongside the change of quantities which should be conserved if the integrator is symplectic (i.e energy)
        self.set_arrays()

        # the number of steps that the integrator has been run for
        self.int_step = 1

    def integration_step(self, t, delta):
        """
        Overloaded method, dependent on the integration scheme utilised.
        At a given step t, uses delta to calculates positions, velocities and acceleration (in some cases)
        :param t: the step at which the calculation is made
        :param delta: the timestep to use at the integration step
        :return: newly calculated positions, velocities (for step t), and in some cases acceleration (for time t+1)
                 Acceleration is returned for Leapfrog2Int and Leapfrog3
        """

        return None, None, None

    def simulation_step(self, t):
        """
        For a given step, calculates the new positions and velocities via the integration step.
        Then, updates the simulation with these values.
        :param t: the step at which the calculation is made
        """

        # check: step t is the same as the expected step int_step.
        # Ensures that when performing an integration_step, they happen at consecutive times
        # For example, integration_step(42) can only be performed if we have previously executed integration_step(41)
        assert self.int_step == t, f"Attempted to integrate with a discontinuous time step. \n" \
                                   f"Step to Integrate: {t}\n" \
                                   f"Expected Step to Integrate: {self.int_step}\n"

        new_positions, new_velocities, acc_tt = self.integration_step(t, self.delta)

        # calculate adaptive delta by using a time-reversible delta
        if self.adaptive:
            # average of adaptive delta at time t and t+1
            reversible_delta = 0.5 * (self.delta + nm.variable_delta(new_positions, new_velocities, adaptive_constant= self.adaptive_constant, delta_lim = self.delta_lim))

            # use reversible delta at time t again to calculate new positions and velocities
            # this ensures that when using adaptive delta, the integrator remains symplectic
            new_positions, new_velocities, acc_tt = self.integration_step(t, delta=reversible_delta)

            # recalculate adaptive timestep
            self.delta = nm.variable_delta(new_positions, new_velocities, adaptive_constant= self.adaptive_constant, delta_lim = self.delta_lim)

        self.update_simulation(t, new_positions, new_velocities, symplectic=True)

        # set the calculated acceleration for the next iteration
        if acc_tt is not None:
            self.acc_t = acc_tt


    def update_historic(self, t):
        """
        Updates the historic arrays of quantities which should be conserved (i.e energy)
        :param t: the step for which the historic quantites are calculated
        """

        assert t != 0

        self.historic_energy[t] = self.nbody.energy
        self.historic_kinetic_energy[t] = self.nbody.kinetic_energy
        self.historic_gpe[t] = self.nbody.gpe
        self.historic_angular_momentum[t,:] = self.nbody.total_angular_momentum

        if self.adaptive:
            self.historic_delta[t] = self.delta

    def update_simulation(self, t, new_positions, new_velocities, symplectic):
        """
        Given newly calculated positions and velocities, updates NBody, alongside historic arrays
        :param t: step at which positions and velocities were calculated
        :param new_positions: newly calculated positions by integrator
        :param new_velocities: newly calculated velocities by integrator
        :param symplectic: True for symplectic integrator,
                           checks that energy,angular momentum.etc... are conserved after the update
        """

        # update the simulation with the calculated position and velocities
        self.nbody.update(new_positions, new_velocities, symplectic=symplectic, tolerance=self.tolerance)

        # add the newly calculated energies and angular momentum (and adaptive delta) to the historic arrays
        if self.store_properties:
            self.update_historic(t)

        # set the newly calculated positions and velocities to the orbit arrays
        self.position_orbit[:, t, :] = new_positions
        self.velocity_orbit[:, t, :] = new_velocities

        # increment the int_step to ensure that integration_step is performed on continuous steps
        self.int_step = t + 1

    def get_orbits(self):
        """
        Performs the integration, by applying the integration step as many times as specified by the initialisation
        If get_orbits has already been executed once, will remove all the previously calculated steps, and calculate new orbits
        """
        # check to see if integration has already been performed
        if self.integrated:
            self.set_arrays()
            self.get_orbits()
        else:
            # if adaptive timestep, we iterate until a limit number of steps, or until the target time is attained
            if self.adaptive:
                while self.int_step < self.steps and self.times[-1] < self.target_time:
                    self.times.append(self.delta + self.times[-1])
                    self.simulation_step(self.int_step)

                # since the limit is set to a larger number than necessary,
                # we shorten all the position, velocity and historic arrays
                # so that they are the same length as the nuber of steps required to integrate until the target time
                assert len(self.times) <= self.steps

                if len(self.times) == self.steps:
                    self.full_run = True

                self.steps = len(self.times)
                self.position_orbit = self.position_orbit[:, :self.steps,:]
                self.velocity_orbit = self.velocity_orbit[:, :self.steps, :]

                if self.store_properties:
                    self.historic_energy = self.historic_energy[:self.steps]
                    self.historic_gpe = self.historic_gpe[:self.steps]
                    self.historic_kinetic_energy = self.historic_kinetic_energy[:self.steps]
                    self.historic_angular_momentum = self.historic_angular_momentum[:self.steps, :]
                    self.historic_delta = self.historic_delta[:self.steps]

            else:
                while self.int_step < self.steps:
                    self.simulation_step(self.int_step)

            # update flag
            self.integrated = True

    def show_orbits(self, animate = False, seed = 42, animation_steps = 1, twodims = True, grid = True):
        """
        Used to visualise the orbits calculated by the integrator
        :param animate: if True, will produce an animated plot of the orbits
        :param seed: used as seed to generate the colours of the plot
        :param animation_steps: number of steps to consider per frame of animation
        :param twodims: if True, will plot the orbit using 2 dimensions (x and y)
        :param grid: if True, will plot in 2 dimensions,
                     whilst displaying information, such as energy, angular momentum, position and velocity change across time
        """

        assert self.integrated, "No orbits were calculated. Run get_orbits() first!"

        plotter = OrbitPlotter(self, seed = seed, animation_steps = animation_steps, twodims = twodims, grid = grid)

        if grid:

            assert self.store_properties, f"Properties were not stored, so grid can't be plotted {grid}"

            plotter.plot_grid()
        else:
            if animate:
                plotter.animate_orbit()
            else:
                plotter.plot_orbit()

    def set_arrays(self):
        """
        Used to initialise all the arrays used by Integrators.
        Generates and initialises arrays for positions, velocities, energies, and angular momentum.
        These arrays contain not only the calculated orbits, but additional information used for plotting or debugging.
        """

        # if adaptive timestep is used, create an array for it
        if self.adaptive:
            self.target_time = self.steps * self.delta
            self.steps = 10**5
            self.delta = nm.variable_delta(self.nbody.positions, self.nbody.velocities, adaptive_constant=self.adaptive_constant, delta_lim = self.delta_lim)
            self.historic_delta = np.zeros(self.steps)
            self.historic_delta[0] = self.delta
            self.times = [0]
            self.full_run = False

        # tensor of shape (n x steps x 3), containing the positions of the n bodies throughout the calculated orbit
        self.position_orbit = np.zeros((self.nbody.n, self.steps, 3))
        self.position_orbit[:, 0, :] = self.nbody.positions

        # tensor of shape (n x steps x 3), containing the velocities of the n bodies throughout the calculated orbit
        self.velocity_orbit = np.zeros((self.nbody.n, self.steps, 3))
        self.velocity_orbit[:, 0, :] = self.nbody.velocities

        # save constants (energies, angular momentum) across time (for plotting purposes)
        if self.store_properties:
            self.historic_energy = np.zeros(self.steps)
            self.historic_energy[0] = self.nbody.energy
            self.historic_kinetic_energy = np.zeros(self.steps)
            self.historic_kinetic_energy[0] = self.nbody.kinetic_energy
            self.historic_gpe = np.zeros(self.steps)
            self.historic_gpe[0] = self.nbody.gpe
            self.historic_angular_momentum = np.zeros(shape=(self.steps, 3))
            self.historic_angular_momentum[0] = self.nbody.total_angular_momentum

        # reset flag (this method is used when the integrator is going to be run)
        self.integrated = False



from copy import deepcopy

import numpy as np

from nbodysim import nmath as nm
from nbodysim.exceptions import *

class NBody:
    """
    Class used to simulate the n-body problem.
    """
    def __init__(self, init_positions, init_velocities, masses, collision_tolerance = 10e-4, escape_tolerance = -1):
        """
        :param init_positions: Python list or numpy array of position vectors for the bodies
        :param init_velocities: Python list or numpy array of velocity vectors for the bodies
        :param masses: Python list or numpy array of the mass of each body
        :param collision_tolerance: maximum distance between 2 bodies allowed before ending simulation
               If None, collisions are not considered.
        :param escape_tolerance: maximum distance away from the centre of mass (COM) allowed before ending simulation
                                 If None, escape_tolerance is automatically calculated
                                 If -1, escape_tolerance is not considered
        """

        # number of bodies in simulation
        self.n = len(init_positions)

        # ensures that input is correctly formatted (in terms of shape & data type)
        self.check_update_input(init_positions, init_velocities, masses)

        self.G = 1#6.67408e-11
        self.collision_tolerance = collision_tolerance

        if escape_tolerance is None:
            # escape tolerance set as 10 times the maximum distance of any body from the COM
            self.escape_tolerance = np.max(nm.ten_norm(self.positions, sqrt = True, axis = 1)) * 10 # can be altered
        else:
            self.escape_tolerance = escape_tolerance

        # total mass of simulation
        self.total_mass = np.sum(self.masses)

        # keep masses as stacked array aswell (convenient for calculations)
        self.stacked_masses = np.vstack(self.masses)

        # calculate COM and linear momentum of initial system
        self.com = self.get_com()
        self.linear_momentum = self.get_lmomentum()
        self.total_linear_momentum = np.sum(self.linear_momentum, axis=0)

        # change to COM coordinates (COM becomes the origin, COM moves at constant velocity)
        self.positions = self.positions - self.com
        self.velocities = self.velocities - (self.total_linear_momentum)/self.total_mass

        # compute a tensor (nxnx3) with entry (i,j) corresponding to the distance vector between the ith and jth body
        self.distances = np.zeros(shape = (self.n, self.n, 3))
        self.get_body_distances()

        # recalculate properties of the system based on COM coordinates
        # linear momentum, total linear momentum, angular momentum, total angular momentum,
        # kinetic energy, gravitational potential energy (GPE) and total energy
        self.linear_momentum = self.get_lmomentum()
        self.total_linear_momentum = np.sum(self.linear_momentum, axis=0)
        self.angular_momentum = self.get_amomentum()
        self.total_angular_momentum = np.sum(self.angular_momentum, axis = 0)
        self.kinetic_energy = 0
        self.gpe = 0
        self.energy = self.get_energy()

        # save the initial energy and momentum values (test for conservation)
        self.first_energy = self.energy
        self.first_angular_momentum = self.total_angular_momentum
        self.first_linear_momentum = self.total_linear_momentum

    def check_update_input(self, init_positions, init_velocities, masses):
        """
        Checks that class initialisation has been done with correct parameters
        Ensures that positions, velocities and masses all have the same number of elements
        Ensure that positions and velocities are matrices of shape (n x 3)
        Lastly, ensures that position, velocity and mass tensors are numpy arrays
        """

        # check: same number of bodies as elements in the position, velocity and mass tensors
        assert (self.n == len(init_velocities) and self.n == len(masses)), \
            f"{self.n} positions given; {len(init_velocities)} velocities given; \
                        {len(masses)} masses given. These 3 quantities must be the same."

        # check: vectors are in R^3
        assert (len(init_positions[0]) == 3 and len(init_velocities[0]) == 3)

        # if parameters are Python lists, change to numpy arrays
        # then, set positions, velocities and masses for the simulation

        if (not isinstance(init_positions, np.ndarray)):
            self.positions = np.array(init_positions)
        else:
            self.positions = init_positions

        if (not isinstance(init_velocities, np.ndarray)):
            self.velocities = np.array(init_velocities)
        else:
            self.velocities = init_velocities

        if (not isinstance(masses, np.ndarray)):
            self.masses = np.array(masses)
        else:
            self.masses = masses

    def get_body_distances(self, positions = None):
        """
        Computes the direction vector from all bodies in an n-body system,
        storing them in a tensor of dimensions (n x n x 3),
        where entry (i,j) contains the direction vector from body i to body j.
        :param positions: the positions of a system of bodies. If None, uses the positions within the simulation.
        :return: if positions is not None, a tensor with the distance vectors between the positions
        """

        if positions is None:
            # itereate over the upper triangular section of the tensor
            for i in range(self.n):
                for j in range(i):
                    # compute distance vector using simulation positions
                    distance_vec = self.positions[j] - self.positions[i]

                    if self.collision_tolerance is not None:
                        # check: no body violates the collision_tolerance distance
                        check_exception(nm.ten_norm(distance_vec, axis = 0, sqrt = True) >= self.collision_tolerance,
                                        BodyCollisionException,
                                        f"A collision occurred. Distance between bodies was {nm.ten_norm(distance_vec, axis=0, sqrt=True)}, but the collision distance is {self.collision_tolerance}.")

                    self.distances[i,j] = distance_vec

                    # direction vector (i,j) is the negative of direction vector (j,i)
                    self.distances[j,i] = -distance_vec
        else:
            # compute the number of bodies & create a corresponding tesnor to store the distance vectors
            n = len(positions)
            distances = np.zeros(shape = (n, n, 3))

            # itereate over the upper triangular section of the tensor
            for i in range(n):
                for j in range(i):
                    # compute distance vector using positions given as arguments
                    distance_vec = positions[j] - positions[i]

                    if self.collision_tolerance is not None:
                        # check: no body violates the collision_tolerance distance
                        if self.collision_tolerance is not None:
                            # check: no body violates the collision_tolerance distance
                            check_exception(nm.ten_norm(distance_vec, axis=0, sqrt=True) >= self.collision_tolerance,
                                            BodyCollisionException,
                                            f"A collision occurred. Distance between bodies was {nm.ten_norm(distance_vec, axis=0, sqrt=True)}, but the collision distance is {self.collision_tolerance}.")

                    distances[i, j] = distance_vec

                    # direction vector (i,j) is the negative of direction vector (j,i)
                    distances[j, i] = -distance_vec

            return distances

    def get_com(self):
        """
        Calculates the COM of the system
        :return: a position vector in 3D for the COM of the system
        """

        com = np.sum(self.stacked_masses * self.positions, axis = 0)/self.total_mass

        return com

    def get_lmomentum(self):
        """
        Calculates the linear momentum (product of mass and velocity) of each body in the system
        :return: an (n x 3) matrix, with the linear momentum of body i at entry i
        """

        lmomentum = self.stacked_masses * self.velocities

        return lmomentum

    def get_amomentum(self):
        """
        Calculates the angular momentum (cross product of position and linear momentum) of each element of the system
        :return: an (n x 3) matrix, with the angular momentum of body i at entry i
        """

        amomentum = nm.mat_cross(self.positions, self.linear_momentum)

        return amomentum

    def get_energy(self):
        """
        Calculates the total energy of the system
        :return: the total energy of the system, given as the sum of kinetic and GPE
        """

        # calculate the total kinetic energy of the system
        kinetic_energy = np.sum(nm.ten_norm(self.linear_momentum, axis = 1, sqrt = False) / (2 * self.masses))
        self.kinetic_energy = kinetic_energy

        # calculate the total GPE of the system
        gpe = 0

        for i in range(self.n):
            for j in range(i):
                gpe += (self.masses[i]*self.masses[j]) / nm.ten_norm(self.distances[i, j], axis = 0)

        gpe *= (-self.G)
        self.gpe = gpe

        return kinetic_energy + gpe

    def get_acceleration(self, positions = None):
        """
        Calculates the acceleration of every particle of a system - solely dependent on position
        :param positions: if None, calculates acceleration based on positions of the system;
                          otherwise, uses positions passed as argument to perform calculation
        :return: an (n x 3) matrix, with each entry i corresponding to the acceleration of o body i
        """

        # calculate the magnitude of the distances between bodies
        if positions is None:
            inv_dist_mag3 = nm.ten_norm(self.distances, axis = 2, sqrt = False)
        else:
            distances = self.get_body_distances(positions = positions)
            inv_dist_mag3 = nm.ten_norm(distances, axis=2, sqrt=False)
        # to avoid division by 0 error, and to avoid unnecessary computational costs,
        # any distance which is 0 is set to infinity, so when reciprocating, the magnitude of the force becomes 0
        inv_dist_mag3[inv_dist_mag3 == 0] = np.inf

        # exponentiate; calculates 1/(magnitude of distance)^3
        inv_dist_mag3 = inv_dist_mag3**(-1.5)

        # tensor, with entry (i,j) giving a unit vector for the direction of the force felt by i due to j
        if positions is None:
            acc_direction = inv_dist_mag3[:,:,np.newaxis] * self.distances
            self.acc_direction = acc_direction
        else:
            acc_direction = inv_dist_mag3[:, :, np.newaxis] * distances

        # calculate the acceleration felt by each body, storing in (n x 3) matrix
        acceleration = np.zeros(shape = (self.n, 3))

        for i in range(self.n):
            for j in range(self.n):
                if i != j:
                    acceleration[i] += self.masses[j] * acc_direction[i, j]

        acceleration *= self.G

        return acceleration

    def conserved_quantity(self, new_value, old_value, tolerance):
        """
        Checks that a quantity is conserved across iterations of the simulation (within a given tolerance)
        :param new_value: newly calculated quantity
        :param old_value: quantity to which new_value is compared to for conservation
        :param tolerance: allowed absolute difference between new_value and old_value
        :return: True if old_value is within tolerance of new_value
        """

        return (np.abs(new_value - old_value) < tolerance).all()

    def update(self, new_positions, new_velocities, symplectic = True, tolerance = 10e-3):
        """
        Updates the simulation, given newly calculated positions and distances.
        :param new_positions: positions to update the system with (as a Python list or numpy array of dimension (n x 3)
        :param new_velocities: velocities to update the system with (as a Python list or numpy array of dimension (n x 3)
        :param symplectic: if symplectic is True, check that calculated quantities (energy, angular momentum, linear momentum) are conserved
        :param tolerance: allowed absolute error for determining conservation of calculated quantities
        """

        # check: same number of bodies as elements in the position and velocity
        assert (self.n == len(new_positions) and self.n == len(new_velocities)), \
            f"{len(new_positions)} positions given; {len(new_velocities)} velocities given; \
                                    These 2 quantities must be the same."

        # check: vectors are in R^3
        assert (len(new_positions[0]) == 3 and len(new_velocities[0]) == 3)

        # check: no bodies gone beyond escape_tolerance
        if self.escape_tolerance != -1:
            check_exception((nm.ten_norm(new_positions, sqrt = True, axis = 1) <= self.escape_tolerance).all(),
                            BodyEscapeException,
                            "A body escaped beyond the allowed distance from the COM.")

        # if positions and velocities have the correct format, update them
        # also recalculate all the distances between the bodies
        self.positions = new_positions
        self.velocities = new_velocities
        self.get_body_distances()

        # calculate new linear momentum (used for kinetic energy and angular momentum calculation)
        new_linear_momentum = self.get_lmomentum()

        new_total_linear_momentum = np.sum(new_linear_momentum, axis = 0)

        self.linear_momentum = new_linear_momentum
        self.total_linear_momentum = new_total_linear_momentum

        # calculate new angular and total angular momentum
        new_angular_momentum = self.get_amomentum()
        new_total_angular_momentum = np.sum(new_angular_momentum, axis=0)

        # calculate new total energy
        new_energy = self.get_energy()

        # calculate new COM
        new_com = self.get_com()

        # check if quantities are conserved if the integration update is meant to be symplectic
        if (symplectic):
            # check: COM at the origin
            check_exception((abs(new_com) <= 10e-10).all(), COMNotConservedException,f"COM should be 0, but is {new_com}")

            # check: total linear momentum conserved
            check_exception(self.conserved_quantity(new_total_linear_momentum, self.first_linear_momentum, tolerance = tolerance),
                            LinearMomentumNotConservedException,
                            f"Total Linear Momentum was NOT conserved after the update.\nInitial Total Linear Momentum: {self.first_linear_momentum}\nCalculated Total Linear Momentum: {new_total_linear_momentum}\n")


            # check: total angular momentum conserved
            check_exception(self.conserved_quantity(new_total_angular_momentum, self.first_angular_momentum, tolerance = tolerance),
                            AngularMomentumNotConservedException,
                            f"Total Angular Momentum was NOT conserved after the update.\nInitial Total Angular Momentum: {self.first_angular_momentum}\nCalculated Total Angular Momentum: {new_total_angular_momentum}\n")

            # check: total energy conserved
            check_exception(self.conserved_quantity(new_energy, self.first_energy, tolerance = tolerance),
                            EnergyNotConservedException,
                            f"Total Energy was NOT conserved after the update.\nInitial Total Energy: {self.first_energy}\nCalculated Total Energy: {new_energy}\n")


        # set the newly calculated values of the system
        self.angular_momentum = new_angular_momentum
        self.total_angular_momentum = new_total_angular_momentum
        self.energy = new_energy

    def copy(self):
        """
        Copies the NBody instance
        """
        return deepcopy(self)

    def __str__(self):
        return f"Bodies: {self.n}\n" + \
               f"Total Mass: {self.total_mass}\n" + \
               f"Centre of Mass: {self.com}\n" + \
               f"Linear Momentum:\n {self.linear_momentum}\n" + \
               f"Total Linear Momentum: {self.total_linear_momentum}\n" + \
               f"Angular Momentum:\n {self.angular_momentum}\n" + \
               f"Total Angular Momentum: {self.total_angular_momentum}\n" + \
               f"Kinetic Energy: {self.kinetic_energy}\n" \
               f"Gravitational Potential Energy: {self.gpe}\n" \
               f"Total Energy: {self.energy}\n"


